from .web import RequestHandler


class UrlFunctionProvider:
    NEW = 'new'

    def update_handler_type(self):
        '''
        Returns the update request handler type object.
        '''
        raise NotImplementedError

    def delete_handler_type(self):
        '''
        Returns the delete request handler type object.
        '''
        raise NotImplementedError

    def listing_handler_type(self):
        '''
        Returns the listing request handler type object.
        '''
        raise NotImplementedError

    def update_url_fn(self):
        def update_url(obj=None):
            update_type = self.update_handler_type()
            if obj:
                id = obj.id
            else:
                id = self.NEW
            return self.reverse_url_for_handler(update_type, id)
        return update_url

    def delete_url_fn(self):
        def delete_url(obj):
            delete_type = self.delete_handler_type()
            return self.reverse_url_for_handler(delete_type, obj.id)
        return delete_url

    def listing_url_fn(self):
        def listing_url(page):
            listing_type = self.listing_handler_type()
            return self.reverse_url_for_handler(listing_type, page)
        return listing_url

    def set_url_functions(self):
        self.set_template_context_key('update_url', self.update_url_fn())
        self.set_template_context_key('delete_url', self.delete_url_fn())
        self.set_template_context_key('listing_url', self.listing_url_fn())


class BaseHandler(RequestHandler):
    OPTION_YES = 'yes'

    @classmethod
    def is_value_true(cls, v):
        v = v.strip().lower()
        if v == cls.OPTION_YES:
            return True
        return False

    def body_argument_is_yes(self, field_name):
        v = self.get_body_argument(field_name, default='No')
        return self.is_value_true(v)

    def body_arguments_are_yes(self, field_name):
        l = self.get_body_arguments(field_name)
        def change_to_bool(v):
            return self.is_value_true(v)
        l = map(change_to_bool, l)
        l = list(l)
        return l

    def set_object_and_render(self, obj):
        '''
        Used to set the object being operated on to make
        the object accessible from the template.
        '''
        self.set_template_context_key('obj', obj)
        return self.template_render()

    def add_flash_message(self, msg):
        '''
        Adds message to show on the next screen which supports showing
        flash messages.

        Child classes should override this.
        '''
        if isinstance(msg, list):
            for m in msg:
                self.add_flash_message(m)
            return
        pass

    def flash_messages(self):
        '''
        Retrieve all the flash messages.

        Child classes should override this.
        '''
        return []


class Listing(BaseHandler):
    @classmethod
    def route(cls):
        return r'/list/([0-9]+)'

    @classmethod
    def template_filename(cls):
        return 'list.html'

    def get_object(self, id):
        '''
        Returns the object given the specified ID.

        Child classes should override this.
        '''
        raise NotImplementedError

    def listing(self, page=None):
        '''
        Returns the list of object IDs to show for the specified page.

        If page is None, it is up to the implementation to interpret what
        that means.

        Child classes should override this.
        '''
        raise NotImplementedError

    def get(self, page=None):
        if not page:
            page = 1
        page = int(page)
        l = self.listing(page)
        def load_entry(e):
            obj = self.get_object(e)
            return obj
        l = map(load_entry, l)
        l = list(l)
        self.set_template_context_key('listing', l)
        self.template_render()


class Update(BaseHandler):
    NEW = 'new'

    @classmethod
    def route(cls):
        return r'/update/(.*)'

    @classmethod
    def template_filename(cls):
        return 'update.html'

    def msg_created(self, obj):
        return 'Create successful.'

    def msg_updated(self, obj):
        return 'Update successful.'

    def new_object(self):
        '''
        Returns a new object.

        Child classes should override this.
        '''
        raise NotImplementedError

    def get_object(self, id):
        '''
        Returns the object given the specified object.

        Child classes should override this.
        '''
        raise NotImplementedError

    def add_object(self, obj):
        '''
        Adds or update the object.

        Child classes should override this.
        '''
        raise NotImplementedError

    def created_message(self, obj):
        '''
        Returns the message for successful operation.

        Arguments:
        obj -- The object which was successfully created.

        Child classes can override this.
        '''
        return 'Create successful.'

    def updated_message(self, obj):
        '''
        Returns the message for successful update operation.

        Arguments:
        obj -- The object which was successful updated.

        Child classes can override this.
        '''
        return 'Update successful.'

    def validate_form(self, obj):
        '''
        Specify the logic for validating the submitted form values.

        The object being passed in is the original object.

        Returns validation error messages; empty list otherwise.

        Child classes can override this.
        '''
        return []

    def validate(self, obj):
        '''
        Validates the object that is about to be saved.
        Returns list of validation error messages; empty list otherwise.

        Child classes can override this.
        '''
        return []

    def parse_into_object(self, obj):
        '''
        Specify the logic for updating the specified object.

        Directly set updates to the specified object.
        The values to use are usually the POSTed values.

        This should be overridden by child classes.
        '''
        raise NotImplementedError

    def perform_post_operation(self, obj):
        '''
        Performs post update operation with the updated object.

        Child classes can override this.
        '''
        pass

    def get(self, id):
        if id == self.NEW:
            obj = self.new_object()
        else:
            obj = self.get_object(id)
        if not obj:
            self.error_not_found()
            return
        self.set_object_and_render(obj)

    def post(self, id):
        if id == self.NEW:
            obj = self.new_object()
        else:
            obj = self.get_object(id)
        if not obj:
            self.error_not_found()
            return
        self.parse_into_object(obj)
        err = self.validate_form(obj)
        if err:
            self.display_message(err)
            self.set_object_and_render(obj)
            return
        err = self.validate(obj)
        if not err:
            self.add_object(obj)
            self.perform_post_operation(obj)
            if id == self.NEW:
                msg = self.created_message(obj)
            else:
                msg = self.updated_message(obj)
            self.display_message(msg)
            url = self.redirect_to_handler(self, obj.id)
            return
        else:
            self.display_message(err)
        self.set_object_and_render(obj)


class Delete(BaseHandler):
    CONFIRM_DELETE = 'confirm_delete'
    POST_DELETE = 'post_delete'

    @classmethod
    def route(cls):
        return r'/delete/(.*)'

    @classmethod
    def update_handler_type(self):
        '''
        Returns the update handler type.

        Child classes should override this.
        '''
        raise NotImplementedError

    @classmethod
    def template_filename(cls):
        return 'delete.html'

    def get_object(self, id):
        '''
        Returns the object given the specified ID.
        '''
        raise NotImplementedError

    def remove_object(self, obj):
        '''
        Logic for removing the object.

        Child classes should override this.
        '''
        raise NotImplementedError

    def set_post_delete_to_template_context(self, v):
        self.set_template_context_key(self.POST_DELETE, v)

    def deleted_message(self, obj):
        '''
        Returns the message if delete is successful.

        Child classes can override this.
        '''
        return 'Deleted.'

    def not_deleted_message(self, obj):
        '''
        Returns the message if delete is not performed.

        Child classes can override this.
        '''
        return 'Not deleted.'

    def perform_post_operation(self, obj):
        '''
        Performs operations after deletion has occurred.

        Child classes can override this.
        '''
        pass

    def get(self, id):
        obj = self.get_object(id)
        if not obj:
            self.error_not_found()
            return
        self.set_post_delete_to_template_context(False)
        self.set_object_and_render(obj)

    def post(self, id):
        obj = self.get_object(id)
        if not obj:
            self.error_not_found()
            return
        resp = self.body_argument_is_yes(self.CONFIRM_DELETE)
        if resp:
            self.remove_object(obj)
            self.perform_post_operation(obj)
            msg = self.deleted_message(obj)
            self.display_message(msg)
            self.set_post_delete_to_template_context(True)
            self.set_object_and_render(obj)
        else:
            msg = self.not_deleted_message(obj)
            self.display_message(msg)
            update_handler_type = self.update_handler_type()
            self.redirect_to_handler(update_handler_type, obj.id)
