from copy import deepcopy
from decimal import Decimal
import datetime
import re
import os.path

try:
    import ujson as json
except ImportError:
    import json


class JSON:
    DATETIME_FORMAT = '%Y-%m-%d %H:%M:%S'
    DATETIME_TWO_DIGIT_PATTERN = r'([0-9]{1,2})'
    DATETIME_FOUR_DIGIT_PATTERN = r'([0-9]{4})'
    DATETIME_PATTERN = r'^%s-%s-%s %s:%s:%s$' % (
        DATETIME_FOUR_DIGIT_PATTERN,
        DATETIME_TWO_DIGIT_PATTERN,
        DATETIME_TWO_DIGIT_PATTERN,
        DATETIME_TWO_DIGIT_PATTERN,
        DATETIME_TWO_DIGIT_PATTERN,
        DATETIME_TWO_DIGIT_PATTERN,
        )

    @classmethod
    def encodable_object(cls, obj):
        '''
        Takes specified object and returns an encodable version.
        '''
        if isinstance(obj, dict):
            for k, v in obj.items():
                obj[k] = cls.encodable_object(v)
        elif isinstance(obj, list):
            length = len(obj)
            for i in range(0, length):
                v = obj[i]
                obj[i] = cls.encodable_object(v)
        elif isinstance(obj, datetime.datetime):
            obj = obj.strftime(cls.DATETIME_FORMAT)
        elif isinstance(obj, Decimal):
            obj = float(obj)
        elif isinstance(obj, JSON):
            obj = obj.to_dict()
            obj = cls.encodable_object(obj)
        return obj

    @classmethod
    def encode(cls, obj):
        '''
        Returns JSON string representation of the specified object.
        '''
        obj = deepcopy(obj)
        obj = cls.encodable_object(obj)
        return json.dumps(obj)

    @classmethod
    def decode(cls, s):
        obj =  json.loads(s)
        if isinstance(obj, dict):
            for k, v in obj.items():
                obj[k] = cls.convert_object(v)
        elif isinstance(obj, list):
            length = len(obj)
            for i in range(0, length):
                v = obj[i]
                obj[i] = cls.convert_object(v)
        else:
            obj = cls.convert_object(obj)
        return obj

    @classmethod
    def is_datetime_string(cls, s):
        if not isinstance(s, str):
            return False
        m = re.match(cls.DATETIME_PATTERN, s)
        if not m:
            return False
        return True
        
    @classmethod
    def convert_to_datetime(cls, s):
        if not isinstance(s, str):
            return None
        return datetime.datetime.strptime(s, cls.DATETIME_FORMAT)

    @classmethod
    def convert_object(cls, obj):
        if isinstance(obj, float):
            return Decimal(obj)
        elif cls.is_datetime_string(obj):
            return cls.convert_to_datetime(obj)
        return obj

    def to_dict(self):
        return self.__dict__


class JSONFile(JSON):
    FILE_EXT = 'json'

    @classmethod
    def decode(cls, s):
        out = cls()
        obj = super().decode(s)
        if isinstance(obj, dict):
            out.__dict__ = obj
        return out

    @classmethod
    def filename(self, name):
        return '%s.%s' % (name, self.FILE_EXT)

    @classmethod
    def read(cls, pathname):
        if not os.path.exists(pathname):
            return None
        with open(pathname, 'r', encoding='utf-8') as f:
            c = f.read()
            return cls.decode(c)
        return None

    def write(self, pathname):
        c = self.encode()
        with open(pathname, 'w', encoding='utf-8') as f:
            f.write(c)

    def encode(self):
        out = super().encode(self)
        return out
