from uuid import uuid4
import getpass
import logging
import sys


class Service:
    '''
    Represents a service.

    This can also be used to represent a system program,
    since a program is also a service.

    All services which inherits from this class
    will be automatically initialized.

    Put the order of the services to initialize in
    the services attribute.

    Put the attribute name of the service which should not be
    initialized in the disable_services attribute.
    '''

    # This can be overridden by child classes to change the service name.
    NAME = 'Core'

    # For the yes-no user prompt
    OPTION_YES = 'y'
    OPTION_NO = 'n'

    INIT_SERVICE_PREFIX = 'init'
    SERVICE_SUFFIX = 'service'

    # Enable to show service initialization progress.
    SHOW_SERVICE_INIT_LOG = False

    # Order of initializing services.
    services = [
    ]

    # List of services to remain uninitialized.
    disable_services = [
    ]

    @classmethod
    def msg(cls, *args, **kwargs):
        '''
        Prints a message to stdout.
        '''
        return print(*args, **kwargs)

    @classmethod
    def main(cls):
        '''
        This should be overridden by child classes.
        '''
        pass

    @classmethod
    def generate_id(cls):
        return str(uuid4())

    def __init__(self):
        self.initialize()

    def new_id(self):
        id = self.generate_id()
        while self.get(id):
            id = self.generate_id()
        return id

    def new(self):
        '''
        Returns a new object of the type the service is representing.

        Child classes can override this.
        '''
        return None

    def get(self, id):
        '''
        Get the object the service is representing.

        Child classes can override this.
        '''
        return None

    def remove(self, obj):
        '''
        Remove the object the service is representing.

        Child classes can override this.
        '''
        pass

    def add(self, obj):
        '''
        Adds or update the object the service is representing.

        Child classes can override this.
        '''
        pass

    def listing(self, page=None):
        '''
        Returns a list of the objects the service is representing.

        Child classes can override this.
        '''
        return []

    def exit(self):
        sys.exit()

    def user_prompt(self, prompt, password=False, yes_no=False,
            yes_no_default=None):
        '''
        If Yes-No mode is enabled, returns either True or False.
        True corresponds to 'Yes' and False corresponds to 'False'.
        '''
        prompt += ' '
        if password:
            ret = getpass.getpass(prompt)
        else:
            ret = input(prompt)
        if yes_no:
            if not yes_no_default:
                yes_no_default = self.OPTION_NO
            ret = ret.strip().lower()
            if ret == '':
                ret = yes_no_default
            if ret == self.OPTION_YES:
                return True
            else:
                return False
        return ret

    def service_attribute_name(self, srv_name):
        '''
        Returns the proper service's attribute name.

        Example:
        self.service_attribute_name('core') => 'core_service'
        '''
        n = '%s_%s' % (
            srv_name,
            self.SERVICE_SUFFIX
            )
        return n

    def init_service_method_name(self, srv_name):
        '''
        Returns the service initialization method name.
        '''
        return '%s_%s_%s' % (
                self.INIT_SERVICE_PREFIX,
                srv_name,
                self.SERVICE_SUFFIX,
                )

    def set_service(self, srv_name, obj):
        n = self.service_attribute_name(srv_name)
        setattr(self, n, obj)

    def has_service_init_method(self, srv_name):
        init_srv_name = self.init_service_method_name(srv_name)
        try:
            init_srv = getattr(self, init_srv_name)
            if not callable(init_srv):
                return False
            return init_srv
        except AttributeError:
            return False

    def initialize(self):
        '''
        Automatically initializes defined services.

        Define services to initialize in the services list.

        Example:
        If you define 'core' in the services list, the 'init_core_service'
        method will be executed. Its return value is then assigned to the
        attribute 'core_service'.
        '''
        for s in self.services:
            init_srv = self.has_service_init_method(s)
            if not init_srv:
                msg = 'Service initialization method must be defined: %s.' \
                     % (s,)
                raise Exception(msg)
            if callable(init_srv) and s not in self.disable_services:
                try:
                    srv_obj = init_srv()
                    srv_obj.initialize()
                    self.set_service(s, srv_obj)
                    if self.SHOW_SERVICE_INIT_LOG:
                        self.msg('Service initialized: %s' % (srv_obj.NAME,))
                except:
                    self.msg('Service initialization error: %s' % (s,))
                    raise
