from tornado.ioloop import IOLoop
from tornado.web import Application as BaseApplication
from tornado.web import RequestHandler as BaseRequestHandler
from tornado.web import URLSpec
from tornado.web import StaticFileHandler
from tornado.httpserver import HTTPServer
from .model import Session
from .json import JSON
import tornado.web


def namespace(name):
    '''
    Decorator. Adds namespace to decorated class.
    '''
    def namespaced_class(cls):
        if cls.__name__ != 'NamespacedClass':
            original_name = '%s.%s' % (
                cls.__module__,
                cls.__name__,
                )
            cls._original_name_ = original_name
        class NamespacedClass(cls, metaclass=Meta):
            @classmethod
            def route(cls):
                r_pattern = '/%s%s'
                r = super().route()
                if isinstance(r, list):
                    out = []
                    for route in r:
                        route = r_pattern % (name, route)
                        out.append(route)
                else:
                    out = r_pattern % (name, r)
                return out

            @classmethod
            def template_filename(cls):
                t = super().template_filename()
                t = '%s/%s' % (name, t)
                return t
        return NamespacedClass
    return namespaced_class


class Meta(type):
    def __repr__(cls):
        route = cls.route()
        out = '<class %s namespaced %s>' % (
            repr(cls._original_name_),
            repr(route),
            )
        return out


class Application(BaseApplication):
    '''
    Represents the web application.

    This utilises a tornado.web.Application object.
    '''

    def __init__(self, request_handlers, static_paths={}, default_host='',
            transforms=None, **settings):
        '''
        Arguments:
        handlers -- list of RequestHandler objects
        static_paths -- object whose keys are the routes themselves and
                        whose values are the pathnames to the directory to be
                        serving static content from (default {})
        default_host -- similar to 'default_host' parameter in
                        tornado.web.Application
        transforms -- similar to 'transforms' parameter in
                      tornado.web.Application
        settings -- similar to 'settings' parameter in tornado.web.Application
        '''
        self.http_server = None
        self.ioloop_instance = None
        self.request_handlers = request_handlers
        self.static_paths = static_paths
        routes = self.generate_route_list()
        super().__init__(routes, default_host, transforms, **settings)

    def generate_route_list(self):
        out_lst = []
        for handler in self.request_handlers:
            route = handler.route()
            if isinstance(route, list):
                for r in route:
                    t = URLSpec(r, handler, name=r)
                    out_lst.append(t)
            elif isinstance(route, str):
                t = URLSpec(route, handler, name=route)
                out_lst.append(t)
        out_lst += self.generate_static_route_list()
        return out_lst

    def generate_static_route_list(self):
        out_lst = []
        for k, v in self.static_paths.items():
            t = URLSpec(k, StaticFileHandler, {'path':v}, name=k)
            out_lst.append(t)
        return out_lst

    def listen(self, port, address='', **kwargs):
        self.http_server = HTTPServer(self, **kwargs)
        return self.http_server.bind(port, address)

    def start(self, num_processes=0):
        '''
        Convenience method for starting the IO loop.
        '''
        self.http_server.start(num_processes)
        self.ioloop_instance = IOLoop.instance()
        return self.ioloop_instance.start()

    def stop(self):
        '''
        Convenience method for stopping the IO loop.
        '''
        self.ioloop_instance.stop()


class RequestHandler(BaseRequestHandler):
    MIME_TYPE_JSON = 'application/json'
    MIME_TYPE_TEXT = 'text/plain'

    @classmethod
    def redirect_url(cls):
        '''
        Returns the URL to redirect to.

        Child classes should override this.
        '''
        return None

    @classmethod
    def route(cls):
        '''
        Route can be a string or a list of strings.
        The route string is identical to how routes are specified in Tornado
        Web Framework.

        Child classes should override this.
        '''
        return []

    @classmethod
    def template_filename(cls):
        '''
        Returns the template filename. This is used as a pathname to load the template files.

        Child classes should override this.
        '''
        return None

    def __init__(self, application, request, **kwargs):
        super().__init__(application, request, **kwargs)
        self.template_context = {}

    def error_not_found(self):
        self.set_status_with_message(404, 'Not found')

    def error_500(self):
        self.set_status_with_message(500, 'Internal Server Error')

    def template_context_key(self, k):
        if k not in self.template_context:
            return None
        return self.template_context[k]

    def redirect(self, url=''):
        if not url:
            url = self.redirect_url()
        super().redirect(url)

    def redirect_to_handler(self, handler_type, *args):
        '''
        Redirects to the specified request handler.
        '''
        url = self.reverse_url_for_handler(handler_type, *args)
        return self.redirect(url)

    def reverse_url_for_handler(self, handler_type, *args):
        r = handler_type.route()
        if isinstance(r, list):
            r = r[0]
        url = self.reverse_url(r, *args)
        return url

    def template_render(self):
        '''
        Similar to tornado.web.RequestHandler#render.
        '''
        f_name = self.template_filename()
        if f_name is None:
            return
        return self.render(f_name, **self.template_context)

    def template_render_string(self, template_name, **kwargs):
        '''
        Similar to tornado.web.RequestHandler#render_string
        '''
        f_name = self.template_filename()
        if f_name is None:
            return
        self.set_flash_messages_to_template_context()
        return self.render(f_name, **self.template_context)

    def set_content_type(self, mime_type, append_charset_utf8=True):
        c = mime_type
        if append_charset_utf8:
            c += '; charset=utf-8'
        return self.set_header('Content-Type', c)

    def set_flash_messages_to_template_context(self):
        flash_messages = self.flash_messages()
        self.set_template_context_key('flash_messages', flash_messages)
        self.clear_flash_messages()

    def set_status_with_message(self, code, message):
        self.set_status(code, message)
        self.set_content_type(self.MIME_TYPE_TEXT)
        self.write(message)

    def set_template_context_key(self, k, v):
        self.template_context[k] = v

    def write_json(self, obj):
        out = JSON.encode(obj)
        self.set_content_type(self.MIME_TYPE_JSON)
        self.write(out)


class RequestHandlerWithFlashMessage(RequestHandler):
    @property
    def session(self):
        '''
        Returns the Session object.

        Child classes should override this.
        '''
        raise NotImplementedError

    def set_flash_messages(self):
        s_obj = self.session
        if not s_obj:
            msgs = []
        else:
            msgs = s_obj.flash_messages()
        self.set_template_context_key(self.FLASH_MESSAGES, msgs)

    def display_message(self, msg):
        '''
        Logic for displaying the message to the user.

        Arguments:
        msg -- A single message in string format or a list of messages.

        Child classes should override this.
        '''
        s_obj = self.session
        s_obj.add_flash_message(msg)
        self.touch_session()

    def touch_session(self):
        '''
        'Touch' the session to keep it updated.

        Child classes should override this.
        '''
        raise NotImplementedError
